#ifndef CDI_TINELIB_H
#define CDI_TINELIB_H

#include "tinetype.h"
#include "errors.h"
#include "cdi.h"
#include "cdieqm.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define CDISCHED_FILE "cdisched.csv"
#define CDIMF_FILE "cdimf.csv"
#define CDIDB_FILE "cdiaddr.csv"
#define CDISRV_FILE "cdiserv.csv"
#define CDI_PARAMETER_DELIMITER ":"
#define CDI_ADDRESS_DELIMITER "."
extern char CdiDBpath[132];
extern char CdiLogPath[132];

#define DEFAULT_QUERY_POLLING_INTERVAL 1000
#define DEFAULT_POLLING_INTERVAL 1000
#define TEMPLATE_RECORD_GAP 100
#define MAX_ERROR_LIST_SIZE 2000
extern int cdiAddrDbSize;
extern int gDefaultQueryPollingRate;

#define HIST_NONE 0
#define HIST_LSN 1
#define HIST_MIN 2
#define HIST_MAX 3
typedef struct cdiHstItem
{
  char dev[32];
  char prp[32];
  int type;
  int fmt;
  int tmr;
  int len;
  int isTmpl;
  struct cdiHstItem *nxt;
} cdiHstItem;
extern cdiHstItem *cdiHstLst;

typedef struct fieldBusStruct
{
  char busType[32];
  char busName[32];
  int  line;
  int  fmt;
  struct fieldBusStruct *nxt;
} FieldBus;
extern FieldBus *gFieldBusList;

typedef struct cdiBusLib
{
  char bus[32];
  HANDLE hlib;
  int flags;
  struct cdiBusLib *nxt;
} CdiBusLib;
extern CdiBusLib *buslibs;
#define BUSLIB_UNLOAD_UNSAFE 0x01


typedef struct extCdiPropertyStruct
{
  char prp[32];
  char tag[16];
  NAME32 *devs;
  int ndevs;
  int fmt;
  int id;
  int isContiguous;
  int isArray;
} ExtCdiProperty;
extern ExtCdiProperty *extCdiPrps;
extern int gNumExtCdiProperties;

typedef struct 
{
  char prp[64];
  char dev[64];
  char tolstr[64];
  float atol;
  float ptol;
  int enabled;
} CdiSchdLstType;
extern CdiSchdLstType *gCdiSchdLst;
extern int sizCdiSchdLst;
extern char *cdiDevStringBuffer;
extern int cdiDevStringBufferSize;

/**
 * \brief Directly executes a synchronous CDI link
 * 
 * Synchronous data acquisition. cdiExecLinkEx() does not complete until the data transfer has completed or 
 * a timeout has ensued. 
 *
 * \param devName is the CDI device name (as registered either via the CDI database or via API call 
 * cdiRegisterDevice().  This can also be a 'group' of CDI devices in the form of 
 * 'device1','device2','device3',... (comma separated) or 'device1' - 'device100' (range specified).
 * \param devProperty is a CDI bus action property (e.g. "SEND", "RECV", "SEND.RECV.ATOM", "RECV.CLBR", etc.)
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the hardware.
 * The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned. Furthermore, the buffer must be large enough to hold the returned data.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the hardware.
 * If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE. etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see cdiAttachLinkEx, DTYPE
 *  
 */
CDI_EXPORT int cdiExecLinkEx(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
);

/**
 * \brief Initiates an asynchronous CDI link
 * 
 * Asynchronous data data acquisition. cdiAttachLinkEx() returns immediately with a positive link index if the 
 * call is properly formed and the targeted device exists. Otherwise,
 * a call to AttachLink() returns a negative completion code which can be interpreted with GetLastLinkError().
 *
 * \param devName is the CDI device name (as registered either via the CDI database or via API call 
 * cdiRegisterDevice().  This can also be a 'group' of CDI devices in the form of 
 * 'device1','device2','device3',... (comma separated) or 'device1' - 'device100' (range specified).
 * \param devProperty is a CDI bus action property (e.g. "SEND", "RECV", "SEND.RECV.ATOM", "RECV.CLBR", etc.)
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the hardware.
 * The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned. Furthermore, the buffer must be large enough to hold the returned data.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the hardware.
 * If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE. etc.).
 * \param pollingRate is the hardware polling interval in milliseconds.
 * \param callback is a caller supplied funtion to be called by CDI when the data acquisition has occurred.
 * It must have the prototype 'void fcn(int id, int cc)', where 'id' will give the CDI link identifier and
 * 'cc' will give the current completion code.
 * \param mode is the requested acquistion mode.  This can be CM_SINGLE (a single asychronous acquisition) or
 * also CM_DATACHANGE or CM_TIMER and can carry other acquistion/notification specifications such as
 * CM_SEEKPATTERN or CM_NODUPLICATES, etc.
 * \param callbackID is a caller supplied id to be returned in the supplied callback function.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see cdiExecLinkEx, DTYPE
 *  
 */
CDI_EXPORT int cdiAttachLinkEx(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
);

CDI_EXPORT int cdiAttachLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
);

CDI_EXPORT int cdiExecLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
);

CDI_EXPORT int SetCdiHome(char *cdiHome);
CDI_EXPORT char *GetCdiHome(void);
CDI_EXPORT void cdilog(char *text,...);
CDI_EXPORT void rotateLogFile(void);
CDI_EXPORT void cdiSetDataByteMask(void *dataMask,int sizeInBytes);
CDI_EXPORT void *cdiGetDataByteMask(void);
CDI_EXPORT char *cdiGetFullDeviceName(char *devName);
CDI_EXPORT char *getCdiLinkName(char *devName,char *lnkName);
CDI_EXPORT int cdiCloseLink(int id);
CDI_EXPORT void initTineCDI(void);
CDI_EXPORT int cdiGetTemplateInstances(char *tmplName,NAME32 *names,int *num);
CDI_EXPORT int cdiGetNumTemplateInstances(char *tmplName);
CDI_EXPORT int cdiGetTemplates(NAME16 *names,int *num);

/**
 * \brief Registers a CDI device
 * 
 * Normally one registers all CDI devices via the CDI database 'cdiaddr.csv'.  However,
 * one can also register hardware devices via this API call.
 *
 * \param devName is the CDI device name to assign to the hardware address specified.
 * \param devNumber is the general CDI device number to assign to the hardware address specified.
 * \param addrMap is an address mapping name to assign to the hardware address specified. This is
 * most useful for busses which deal with memory mapping (such as PLCs).
 * \param devDesc is a description of the device being registered.
 * \param cdiLine is the local hardware bus line on which to issue hardware telegrams.
 * \param cdiIndex is the hardware index on the CDI line specified (i.e. the device number on the line
 * given).
 * \param addrBase is an array of up to 16 integer values defining the address registers.  Most busses
 * take integer address values.  If not, the the addrMap parameter can be used to specify the address.
 * \param addrLen gives the number of integer values being passed in the addrBase (the size of the 
 * integer array).
 * \param addrParam is an array of up to 16 integer values defining additional register parameters
 * used to define the device.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *  
 */
CDI_EXPORT int cdiRegisterDevice(
  char *devName, int devNumber, char *addrMap, char *devDesc,
  int cdiLine, int cdiIndex, int *addrBase, int addrLen,int *addrParam
);
CDI_EXPORT int readCdiDataBase(void);
CDI_EXPORT int readCdiManifest(void);
CDI_EXPORT int readCdiSchedulingInfo(void);
CDI_EXPORT short getCdiFormat(short tineFormat);
CDI_EXPORT short getTineFormat(short cdiFormat);

/**
 * \brief initializes CDI
 * 
 * When statically linking to the CDI library, this routine must be called so as to 
 * initialize the CDI subsystem and read the local manifest and database files.
 * Dynamically linked libraries will automatically call this routine during the
 * prolog phase which the library is brought into memory.
 * 
 * \return 0 upon success 
 */
CDI_EXPORT int cdiInitialize(void);
extern int gGenerateAlarms;
/**
 * \brief Sets the default polling interval for local queries
 * 
 * The local history subsystem and/or the local alarm server subsystem may or
 * may not access CDI devices depending on the local server setup configuration.
 * If CDI devices are accessed in this way, the internal polling interval for these
 * kinds of access can be determined via this call.
 *
 * \param value is the polling interval in milliseconds used to refresh information
 * for the local history and/or local alarm systems.  (default: 1000 milliseconds).
 *
 */
CDI_EXPORT void SetDefaultQueryPollingInterval(int value);

/**
 * \brief Determines whether or not CDI will generate bus error alarms.
 * 
 * If CDI calls to ExecLink() (or cdiExecLinkEx()) or AttachLink() (or cdiAttachLinkEx())
 * determine a bus error during the hardware access, a TINE alarm will be set
 * if the value given here is TRUE, otherwise not.
 *
 * \param value determines whether CDI will set TINE alarms when a bus error
 * is discovered. (default: TRUE)
 *
 */
CDI_EXPORT void cdiSetGenerateAlarms(int value);

/**
 * \brief Sets the number of bus io errors which are tolerated before setting a hardware error alarm
 * 
 * \param value gives the number of consecutive io errors to tolerate
 * (default: 0)
 *
 */
CDI_EXPORT void cdiSetHardwareAlarmThreshold(int value);
extern int gIoErrorThreshold;

/**
 * \brief Gets the number of bus io errors which are tolerated before setting a hardware error alarm
 * 
 * \return the number of consecutive io errors to tolerate
 * (default: 0)
 *
 */
CDI_EXPORT int cdiGetHardwareAlarmThreshold(void);

/**
 * \brief Sets the exported TINE device server name to the value given.
 * 
 * If CDI initializes a TINE hardware server, it will exort itself under the
 * name given.
 *
 * \param expName is the desired export name of the CDI hardware server.
 * (default: \<fecname\>.CDI)
 *
 */
CDI_EXPORT int cdiSetExportName(char *expName);

short getBaseFormat(short tineFormat,int *needsErrList,int *needsNameList);
short getCdiBaseFormat(short tineFormat);
int cdiSystemInitialization(void);
void fixCdiRepository(void);
void remapCdiDeviceAndProperty(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int dbCallHndlr(int pid,char *devName,char *tag,DTYPE *dout,DTYPE *din,short access);
char *cdiGetFieldBusName(char *busType,int line);
int cdiHasAlarms(void);
int isCdiBusAccessProperty(char *prp);
float getFloatAtIndex(BYTE *dat,int idx,int fmt);
int cdiGetAllowedNets(void);
int addBusLib(char *bus,HANDLE *hlib,int flags);
void rmvBusLibs(void);

#ifndef ccerr
# define ccerr(e) {cc = e; goto err; }
#endif

#define CDI_MAJOR_VERSION 2
#define CDI_MINOR_VERSION 1
#define CDI_VERSION_REVISION 1

/**
 * \brief Returns the CDI library version as byte array
 */
CDI_EXPORT BYTE *cdiVersion(void);

/**
 * \brief Returns the CDI library version as string
 */
CDI_EXPORT char *cdiGetVersionAsString(void);

/**
 * \brief Turns on/off remote management 
 *
 * If remote management is turned on, then the
 * stock properties 'SRVINIT', 'SRVEXIT', and 'SRVRESET'
 * are allowed to re-initialize, exit, or reset the 
 * server's equiment module.  Specifically, 'SRVEXIT' will
 * issue an 'exit(code)' at the process level, where 'code'
 * is the input given in the call to 'SRVEXIT'.
 * 
 * \param value is the desired setting (default = FALSE).
 *
 */
CDI_EXPORT void cdiSetAllowRemoteManagement(int value);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CDI_TINELIB_H */

